import io
import asyncio
from typing import Optional, Union, Generator, AsyncGenerator, List
from pydantic import BaseModel, Field, ConfigDict
from .ai_config import AIConfig
from .models_toolkit import ModelsToolkit
from .base_toolkit_model import OpenAIMessage


class TextResponse(BaseModel):
    text: str = Field(
        default="",
        description="Text response generated by the model.",
    )


class TextStreamingResponse(BaseModel):
    text_response: bool = Field(
        default=False,
        description="Indicates if the response is a text response.",
    )


class ImageResponse(BaseModel):
    image_description_to_generate: str = Field(
        default="",
        description="Description of the generated image.",
    )


class AudioResponse(BaseModel):
    audio_description_to_generate: str = Field(
        default="",
        description="Description of the generated audio.",
    )


class TextWithImageResponse(BaseModel):
    text: str = Field(
        default="",
        description="Text response generated by the model.",
    )
    image_description_to_generate: str = Field(
        default="",
        description="Description of the generated image.",
    )


class TextWithImageStreamingResponse(BaseModel):
    text_generation_needed: bool = Field(
        default=False,
        description="Indicates if text generation is needed.",
    )
    image_description_to_generate: str = Field(
        default="",
        description="Description of the generated image.",
    )


class OmniModelOutputType(BaseModel):
    output_type: Union[
        TextResponse,
        ImageResponse,
        AudioResponse,
        TextWithImageResponse,
    ] = Field(
        description="Type of output expected from the model.",
    )


class OmniModelStreamingOutputType(BaseModel):
    output_type: Union[
        TextStreamingResponse,
        ImageResponse,
        AudioResponse,
        TextWithImageStreamingResponse,
    ] = Field(
        description="Type of output expected from the model during streaming.",
    )


class OmniModelOutput(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    total_text: Optional[str] = Field(
        default=None,
        description="Text response from the model.",
    )
    text_new_chunk: Optional[str] = Field(
        default=None,
        description="New chunk of text response from the model.",
    )
    image_url: Optional[str] = Field(
        default=None,
        description="Generated image URL response from the model.",
    )
    audio_bytes: Optional[io.BytesIO] = Field(
        default=None,
        description="In-memory audio bytes response from the model.",
    )
    total_price: Optional[float] = Field(
        default=None,
        description="Total price of the model response based on input and output.",
    )


class OmniModel:
    def __init__(
        self, openai_api_key: Optional[str] = None, ai_config: Optional[AIConfig] = None
    ):
        self.ai_config = ai_config
        self.modkit = ModelsToolkit(openai_api_key=openai_api_key, ai_config=ai_config)

    def _compose_user_input(
        self,
        user_input: Optional[str] = None,
        image_description: Optional[str] = None,
        audio_description: Optional[str] = None,
    ) -> str:
        """
        Compose the user input by combining text, image description, and audio description.
        """
        if user_input is None:
            user_input = ""

        if image_description is not None:
            user_input += f" {image_description}"

        if audio_description is not None:
            user_input += f" {audio_description}"

        return user_input.strip()

    def _get_user_input(
        self,
        user_input: Optional[str] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> str:
        image_description = None
        if in_memory_image_stream is not None:
            image_description_object = self.modkit.vision_model.run_default(
                in_memory_image_stream=in_memory_image_stream,
            )
            image_description = str(image_description_object)

        audio_description = None
        if in_memory_audio_stream is not None:
            audio_description_object = self.modkit.audio_recognition_model.run_default(
                in_memory_audio_stream=in_memory_audio_stream,
            )
            audio_description = str(audio_description_object)

        return self._compose_user_input(
            user_input=user_input,
            image_description=image_description,
            audio_description=audio_description,
        )

    async def _aget_user_input(
        self,
        user_input: Optional[str] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> str:
        image_description = None
        if in_memory_image_stream is not None:
            image_description_object = await self.modkit.vision_model.arun_default(
                in_memory_image_stream=in_memory_image_stream,
            )
            image_description = str(image_description_object)
        audio_description = None
        if in_memory_audio_stream is not None:
            audio_description_object = (
                await self.modkit.audio_recognition_model.arun_default(
                    in_memory_audio_stream=in_memory_audio_stream,
                )
            )
            audio_description = str(audio_description_object)
        return self._compose_user_input(
            user_input=user_input,
            image_description=image_description,
            audio_description=audio_description,
        )

    def run(
        self,
        user_input: Optional[str] = None,
        system_prompt: Optional[str] = None,
        communication_history: Optional[List[OpenAIMessage]] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> OmniModelOutput:
        """
        Run the OmniModel with the provided inputs and return the output.
        """
        user_input = self._get_user_input(
            user_input=user_input,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

        # Determine the output type based on the input data
        output_type_model = self.modkit.text_model.run(
            system_prompt=system_prompt,
            pydantic_model=OmniModelOutputType,
            user_input=user_input,
            communication_history=communication_history,
        )
        output_type = output_type_model.output_type

        # Process the input data based on the output type
        if isinstance(output_type, ImageResponse):
            image_response = self.modkit.image_generation_model.run_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            result = OmniModelOutput(image_url=image_response.image_url)
        elif isinstance(output_type, AudioResponse):
            audio_response = self.modkit.audio_generation_model.run_default(
                system_prompt=system_prompt,
                user_input=output_type.audio_description_to_generate,
                communication_history=communication_history,
            )
            result = OmniModelOutput(audio_bytes=audio_response.audio_bytes)
        elif isinstance(output_type, TextWithImageResponse):
            image_response = self.modkit.image_generation_model.run_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            result = OmniModelOutput(
                total_text=output_type.text,
                image_url=image_response.image_url,
            )
        elif isinstance(output_type, TextResponse):
            result = OmniModelOutput(total_text=output_type.text)
        else:
            raise ValueError("Unexpected output type received from the model.")

        return self.inject_price(
            output=result,
            user_input=user_input,
            system_prompt=system_prompt,
            communication_history=communication_history,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

    async def arun(
        self,
        user_input: Optional[str] = None,
        system_prompt: Optional[str] = None,
        communication_history: Optional[List[OpenAIMessage]] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> OmniModelOutput:
        """
        Asynchronously run the OmniModel with the provided inputs and return the output.
        """
        user_input = await self._aget_user_input(
            user_input=user_input,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

        # Determine the output type based on the input data
        output_type_model = await self.modkit.text_model.arun(
            system_prompt=system_prompt,
            pydantic_model=OmniModelOutputType,
            user_input=user_input,
            communication_history=communication_history,
        )
        output_type = output_type_model.output_type

        # Process the input data based on the output type
        if isinstance(output_type, ImageResponse):
            image_response = await self.modkit.image_generation_model.arun_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            result = OmniModelOutput(image_url=image_response.image_url)
        elif isinstance(output_type, AudioResponse):
            audio_response = await self.modkit.audio_generation_model.arun_default(
                system_prompt=system_prompt,
                user_input=output_type.audio_description_to_generate,
                communication_history=communication_history,
            )
            result = OmniModelOutput(audio_bytes=audio_response.audio_bytes)
        elif isinstance(output_type, TextWithImageResponse):
            image_response = await self.modkit.image_generation_model.arun_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            result = OmniModelOutput(
                total_text=output_type.text,
                image_url=image_response.image_url,
            )
        elif isinstance(output_type, TextResponse):
            result = OmniModelOutput(total_text=output_type.text)
        else:
            raise ValueError("Unexpected output type received from the model.")
        return self.inject_price(
            output=result,
            user_input=user_input,
            system_prompt=system_prompt,
            communication_history=communication_history,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

    def stream(
        self,
        user_input: Optional[str] = None,
        system_prompt: Optional[str] = None,
        communication_history: Optional[List[OpenAIMessage]] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> Generator[OmniModelOutput, None, None]:
        """
        Run the OmniModel with the provided inputs and return the output.
        """
        user_input = self._get_user_input(
            user_input=user_input,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

        # Determine the output type based on the input data
        output_type_model = self.modkit.text_model.run(
            system_prompt=system_prompt,
            pydantic_model=OmniModelStreamingOutputType,
            user_input=user_input,
            communication_history=communication_history,
        )
        output_type = output_type_model.output_type

        # Process the input data based on the output type
        if isinstance(output_type, ImageResponse):
            image_response = self.modkit.image_generation_model.run_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            final_response = OmniModelOutput(image_url=image_response.image_url)
        elif isinstance(output_type, AudioResponse):
            audio_response = self.modkit.audio_generation_model.run_default(
                system_prompt=system_prompt,
                user_input=output_type.audio_description_to_generate,
                communication_history=communication_history,
            )
            final_response = OmniModelOutput(audio_bytes=audio_response.audio_bytes)
        elif isinstance(output_type, TextWithImageStreamingResponse):
            image_response = self.modkit.image_generation_model.run_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            text_response = self.modkit.text_model.stream_default(
                system_prompt=system_prompt,
                user_input=user_input,
                communication_history=communication_history,
            )
            # No point in streaming text with image synchronously,
            # as the image generation is a long-running task
            # (longer than text generation).
            final_response = OmniModelOutput(
                total_text=text_response,
                image_url=image_response.image_url,
            )
        elif isinstance(output_type, TextStreamingResponse):
            total_text = ""
            for chunk in self.modkit.text_model.stream_default(
                system_prompt=system_prompt,
                user_input=user_input,
                communication_history=communication_history,
            ):
                total_text += chunk.text_chunk
                yield OmniModelOutput(
                    total_text=total_text,
                    text_new_chunk=chunk.text_chunk,
                )
            final_response = OmniModelOutput(total_text=total_text, text_new_chunk="")
        else:
            raise ValueError("Unexpected output type received from the model.")
        return self.inject_price(
            output=final_response,
            user_input=user_input,
            system_prompt=system_prompt,
            communication_history=communication_history,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

    async def astream(
        self,
        user_input: Optional[str] = None,
        system_prompt: Optional[str] = None,
        communication_history: Optional[List[OpenAIMessage]] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> AsyncGenerator[OmniModelOutput, None]:
        """
        Asynchronously run the OmniModel with the provided inputs and return the output.
        """
        user_input = await self._aget_user_input(
            user_input=user_input,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

        # Determine the output type based on the input data
        output_type_model = await self.modkit.text_model.arun(
            system_prompt=system_prompt,
            pydantic_model=OmniModelStreamingOutputType,
            user_input=user_input,
            communication_history=communication_history,
        )
        output_type = output_type_model.output_type

        # Process the input data based on the output type
        if isinstance(output_type, ImageResponse):
            image_response = await self.modkit.image_generation_model.arun_default(
                system_prompt=system_prompt,
                user_input=output_type.image_description_to_generate,
                communication_history=communication_history,
            )
            final_response = OmniModelOutput(image_url=image_response.image_url)
        elif isinstance(output_type, AudioResponse):
            audio_response = await self.modkit.audio_generation_model.arun_default(
                system_prompt=system_prompt,
                user_input=output_type.audio_description_to_generate,
                communication_history=communication_history,
            )
            final_response = OmniModelOutput(audio_bytes=audio_response.audio_bytes)
        elif isinstance(output_type, TextWithImageStreamingResponse):
            image_response_task = asyncio.create_task(
                self.modkit.image_generation_model.arun_default(
                    system_prompt=system_prompt,
                    user_input=output_type.image_description_to_generate,
                    communication_history=communication_history,
                )
            )
            total_text = ""
            async for chunk in self.modkit.text_model.astream_default(
                system_prompt=system_prompt,
                user_input=user_input,
                communication_history=communication_history,
            ):
                if chunk.text_chunk:
                    total_text += chunk.text_chunk
                    yield OmniModelOutput(
                        total_text=total_text,
                        text_new_chunk=chunk.text_chunk,
                    )
            image_response = await image_response_task
            final_response = OmniModelOutput(
                total_text=total_text,
                image_url=image_response.image_url,
            )
        elif isinstance(output_type, TextResponse):
            total_text = ""
            async for chunk in self.modkit.text_model.astream_default(
                system_prompt=system_prompt,
                user_input=user_input,
                communication_history=communication_history,
            ):
                total_text += chunk.text_chunk
                yield OmniModelOutput(
                    total_text=total_text,
                    text_new_chunk=chunk.text_chunk,
                )
            final_response = OmniModelOutput(total_text=total_text, text_new_chunk="")
        else:
            raise ValueError("Unexpected output type received from the model.")
        yield self.inject_price(
            output=final_response,
            user_input=user_input,
            system_prompt=system_prompt,
            communication_history=communication_history,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )

    def get_price(
        self,
        output: OmniModelOutput,
        user_input: Optional[str] = None,
        system_prompt: Optional[str] = None,
        communication_history: Optional[List[OpenAIMessage]] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> float:
        """
        Get the price of the model based on input and output text, image, and audio.
        """
        total_input_text = (
            (user_input or "")
            + (system_prompt or "")
            + "\n".join([msg["text"] for msg in communication_history or []])
        )
        return self.modkit.get_price(
            input_text=total_input_text,
            output_text=output.total_text,
            input_image=in_memory_image_stream,
            output_image_url=output.image_url,
            input_audio=in_memory_audio_stream,
            output_audio=output.audio_bytes,
        )

    def inject_price(
        self,
        output: OmniModelOutput,
        user_input: Optional[str] = None,
        system_prompt: Optional[str] = None,
        communication_history: Optional[List[OpenAIMessage]] = None,
        in_memory_image_stream: Optional[io.BytesIO] = None,
        in_memory_audio_stream: Optional[io.BytesIO] = None,
    ) -> OmniModelOutput:
        """
        Inject the price into the OmniModelOutput based on the input and output.
        """
        output.total_price = self.get_price(
            output=output,
            user_input=user_input,
            system_prompt=system_prompt,
            communication_history=communication_history,
            in_memory_image_stream=in_memory_image_stream,
            in_memory_audio_stream=in_memory_audio_stream,
        )
        return output
